@@ -539,7 +539,7 @@ mount_cb (DBusMessage *reply,
                                         data->callback,
                                         data->user_data,
                                         NULL);
-  g_simple_async_result_complete (simple);
+  g_simple_async_result_complete_in_idle (simple);
   g_object_unref (simple);
 
   g_object_unref (data->object);
